##############################################################################
#
#    Copyright (C) 2010 Almacom (Thailand) Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv,fields,expression
import netsvc
import time
import datetime
import re
from tools.translate import _
from mx import DateTime

class mrp_workcenter(osv.osv):
    _inherit="mrp.workcenter"

    # make workcenters searchable by code
    def name_search(self,cr,uid,name="",args=[],operator='ilike',context={},limit=80):
        if not name:
            ids=self.search(cr,uid,args,limit=limit,context=context)
        else:
            ids=self.search(cr,uid,[("code","=",name)]+args,limit=limit,context=context)
            if not ids:
                ids=self.search(cr,uid,[("code",operator,name)]+args,limit=limit,context=context)
                ids+=self.search(cr,uid,[("name",operator,name)]+args,limit=limit,context=context)
        res=self.name_get(cr,uid,ids,context)
        return res
mrp_workcenter()

class mrp_routing(osv.osv):
    _inherit="mrp.routing"

    # make routings searchable by code
    def name_search(self,cr,uid,name="",args=[],operator='ilike',context={},limit=80):
        if not name:
            ids=self.search(cr,uid,args,limit=limit,context=context)
        else:
            ids=self.search(cr,uid,[("code","=",name)]+args,limit=limit,context=context)
            if not ids:
                ids=self.search(cr,uid,[("code",operator,name)]+args,limit=limit,context=context)
                ids+=self.search(cr,uid,[("name",operator,name)]+args,limit=limit,context=context)
        res=self.name_get(cr,uid,ids,context)
        return res
mrp_routing()

class mrp_routing_workcenter(osv.osv):
    _inherit="mrp.routing.workcenter"

    def _get_hours(self,cr,uid,ids,field_name,arg,context):
        vals={}
        for rwc in self.browse(cr,uid,ids):
            vals[rwc.id]=rwc.minute_nbr/60.0
        return vals

    # enter routing workcenter usage in minutes instead of hours
    _columns={
        'minute_nbr': fields.float('Number of Minutes', required=True),
        "hour_nbr": fields.function(_get_hours,method=True,string="Number of Hours",type="float"),
    }
mrp_routing_workcenter()

class mrp_production_workcenter_line(osv.osv):
    _inherit="mrp.production.workcenter.line"

    def _get_minute(self,cr,uid,ids,field_name,arg,context):
        vals={}
        for line in self.browse(cr,uid,ids):
            vals[line.id]=line.hour*60.0
        return vals

    def _set_minute(self,cr,uid,ids,name,value,arg,context):
        self.write(cr,uid,ids,{"hour":value/60.0})
        return True

    # enter production order workcenter usage in minutes instead of hours
    _columns={
        "minute": fields.function(_get_minute,method=True,string="Nbr of minute",type="float",fnct_inv=_set_minute),
        'hour': fields.float('Nbr of hour',digits=(16,2),readonly=True),
    }
mrp_production_workcenter_line()

class stock_location(osv.osv):
    _inherit="stock.location"
    # add location type for scrapped products
    _columns={
        'usage': fields.selection([('supplier', 'Supplier Location'), ('view', 'View'), ('internal', 'Internal Location'), ('customer', 'Customer Location'), ('inventory', 'Inventory'), ('procurement', 'Procurement'), ('production', 'Production'),('scrap','Scrapped Products')], 'Location Type', required=True),
    }
stock_location()

class ac_production(osv.osv):
    _name="ac.production"

    def _fin_qty_total(self,cr,uid,ids,name,arg,context={}):
        vals={}
        for order in self.browse(cr,uid,ids):
            qty=0.0
            for fin in order.fin_products:
                qty+=fin.qty
            vals[order.id]=qty
        return vals

    def _get_time(self,cr,uid,ids,name,args,context={}):
        vals_={}
        for order in self.browse(cr,uid,ids):
            vals={
                "time_planned": 0.0,
                "time_actual": 0.0,
            }
            for op in order.operations:
                if op.time_planned:
                    vals["time_planned"]+=op.time_planned
                if op.time_actual:
                    vals["time_actual"]+=op.time_actual
            vals_[order.id]=vals
        return vals_

    def _get_time_start(self,cr,uid,ids,name,arg,context={}):
        print "PROD _get_time_start",ids
        vals={}
        for order in self.browse(cr,uid,ids):
            if order.operations:
                time=min([op.time_start for op in order.operations])
            else:
                time=order.date_planned+" 08:00:00"
            vals[order.id]=time
        return vals

    def _get_time_stop(self,cr,uid,ids,name,arg,context={}):
        print "PROD _get_time_stop",ids
        vals={}
        for order in self.browse(cr,uid,ids):
            if order.operations:
                time=max([op.time_stop for op in order.operations])
            else:
                time=order.date_planned+" 17:00:00"
            vals[order.id]=time
        return vals

    _columns={
        "name": fields.char("Reference",size=64,required=True,select=1),
        "product_id": fields.many2one("product.product","Product",required=True,select=2,domain=[("mrp_type","=","part")]),
        "qty": fields.float("Quantity",required=True),
        "uom_id": fields.many2one("product.uom","UoM",required=True),
        "cons_products": fields.one2many("ac.production.consumed.product","order_id","Consumed Products"),
        "cons_retrieved_by": fields.many2one("res.users","Retrieved By"),
        "cons_date": fields.date("Date Retrieved"),
        "cons_approved_by": fields.many2one("res.users","Approved By"),
        "cons_checked_by": fields.many2one("res.users","Checked By"),
        "operations": fields.one2many("ac.workcenter.operation","production_id","Workcenter Operations"),
        "steps": fields.one2many("ac.production.step","order_id","Production Steps"),
        "fin_date": fields.date("Date Checked"),
        "fin_qty_total": fields.function(_fin_qty_total,method=True,type="float",string="Total Qty"),
        "fin_qty_good": fields.float("Good Qty"),
        "fin_qty_bad": fields.float("Bad Qty"),
        "fin_checked_by": fields.many2one("res.users","Checked By"),
        "fin_products": fields.one2many("ac.production.finished.product","order_id","Finished Products"),
        "fin_submitted_by": fields.many2one("res.users","Submitted By"),
        "fin_submitted_date": fields.date("Submitted Date"),
        "fin_received_by": fields.many2one("res.users","Received By"),
        "sched_products": fields.one2many("ac.scheduled.product","production_id","Scheduled Products",readonly=True),
        "state": fields.selection([("draft","Draft"),("confirmed","Confirmed"),("in_progress","In Progress"),("done","Done"),("canceled","Canceled")],"Status",readonly=True,required=True),
        "location_src_id": fields.many2one("stock.location","Raw Material Location",required=True,domain=[('usage','=','internal')]),
        "location_dest_id": fields.many2one("stock.location","Finished Parts Location",required=True,domain=[('usage','=','internal')]),
        "bom_id": fields.many2one("mrp.bom","Bill of Material"),
        "routing_id": fields.many2one("mrp.routing","Routing"),
        "date_planned": fields.date("Scheduled Date",required=True,select=1),
        "time_planned": fields.function(_get_time,method=True,type="float",string="Planned Time",multi="time"),
        "time_actual": fields.function(_get_time,method=True,type="float",string="Actual Time",multi="time"),
        "origin": fields.char("Origin",size=64,select=1),
        "time_start": fields.function(_get_time_start,method=True,type="datetime",string="Start Time"),
        "time_stop": fields.function(_get_time_stop,method=True,type="datetime",string="Stop Time"),
        "fin_move_total_id": fields.many2one("stock.move","Total Finished Parts Move"),
        "fin_move_good_id": fields.many2one("stock.move","Good Finished Parts Move"),
        "fin_move_bad_id": fields.many2one("stock.move","Bad Finished Parts Move"),
    }
    _defaults={
        "name": lambda self,cr,uid,context: self.pool.get('ir.sequence').get(cr,uid,'mrp.production') or '/',
        "state": lambda *a: "draft",
        "date_planned": lambda *a: time.strftime("%Y-%m-%d"),
    }
    _order="date_planned"

    def onchange_product(self,cr,uid,ids,prod_id,context={}):
        if not prod_id:
            return {}
        prod=self.pool.get("product.product").browse(cr,uid,prod_id)
        vals={
            "uom_id": prod.uom_id.id,
            "qty": 1.0,
        }
        return {"value": vals}

    def onchange_fin_qty_good(self,cr,uid,ids,fin_qty_good,fin_qty_total):
        if not fin_qty_good:
            return {}
        vals={
            "fin_qty_bad": fin_qty_total-fin_qty_good,
        }
        return {"value":vals}

    def onchange_fin_qty_bad(self,cr,uid,ids,fin_qty_bad,fin_qty_total):
        if not fin_qty_bad:
            return {}
        vals={
            "fin_qty_good": fin_qty_total-fin_qty_bad,
        }
        return {"value":vals}

    def btn_compute(self,cr,uid,ids,properties=[],context={}):
        results = []
        for order in self.browse(cr, uid, ids):
            cr.execute('delete from ac_scheduled_product where production_id=%s', (order.id,))
            cr.execute('delete from ac_workcenter_operation where production_id=%s', (order.id,))
            bom_point = order.bom_id
            bom_id = order.bom_id.id
            if not bom_point:
                bom_id = self.pool.get('mrp.bom')._bom_find(cr, uid, order.product_id.id, order.uom_id.id, properties)
                if bom_id:
                    bom_point = self.pool.get('mrp.bom').browse(cr, uid, bom_id)
                    routing_id = bom_point.routing_id.id or False
                    self.write(cr, uid, [order.id], {'bom_id': bom_id, 'routing_id': routing_id})

            if not bom_id:
                raise osv.except_osv(_('Error'), _("Couldn't find bill of material for product"))

            factor = order.qty * order.uom_id.factor / bom_point.product_uom.factor
            res = self.pool.get('mrp.bom')._bom_explode(cr, uid, bom_point, factor / bom_point.product_qty, properties)
            results = res[0]
            results2 = res[1]
            for line in results:
                vals={
                    "production_id": order.id,
                    "product_id": line["product_id"],
                    "qty": line["product_qty"],
                    "uom_id": line["product_uom"],
                }
                self.pool.get('ac.scheduled.product').create(cr, uid, vals)
            for line in results2:
                vals={
                    "production_id": order.id,
                    "sequence": line["sequence"],
                    "name": line["name"],
                    "workcenter_id": line["workcenter_id"],
                    "time_planned": line["hour"]*60.0,
                    "qty": order.qty,
                    "date": order.date_planned,
                }
                self.pool.get('ac.workcenter.operation').create(cr, uid, vals)
        return True

    def wkf_draft(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"draft"})
        for order in self.browse(cr,uid,ids):
            cr.execute('delete from ac_production_consumed_product where order_id=%s', (order.id,))
            cr.execute('delete from ac_production_finished_product where order_id=%s', (order.id,))
        return True

    def wkf_confirmed(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"confirmed"})
        for order in self.browse(cr,uid,ids):
            if not order.routing_id or not order.routing_id.location_id:
                raise osv.except_osv("Error","Missing routing production location")
            for prod in order.sched_products:
                vals={
                    "name": "CONS_TOT: "+order.name,
                    "product_id": prod.product_id.id,
                    "product_qty": prod.qty,
                    "product_uom": prod.uom_id.id,
                    "location_id": order.location_src_id.id,
                    "location_dest_id": order.routing_id.location_id.id,
                    "state": "confirmed",
                    "production_id": order.id,
                }
                move_id=self.pool.get("stock.move").create(cr,uid,vals)
                vals={
                    "order_id": order.id,
                    "date": None,
                    "product_id": prod.product_id.id,
                    "qty_total": prod.qty,
                    "qty_good": prod.qty,
                    "qty_bad": 0.0,
                    "uom_id": prod.uom_id.id,
                    "move_total_id": move_id,
                }
                cons_id=self.pool.get("ac.production.consumed.product").create(cr,uid,vals)
                proc_id = self.pool.get('mrp.procurement').create(cr, uid, {
                    'name': (order.origin or '').split(':')[0] + ':' + order.name,
                    'origin': (order.origin or '').split(':')[0] + ':' + order.name,
                    'date_planned': order.date_planned,
                    'product_id': prod.product_id.id,
                    'product_qty': prod.qty,
                    'product_uom': prod.uom_id.id,
                    'location_id': order.location_src_id.id,
                    'procure_method': prod.product_id.procure_method,
                    'move_id': move_id,
                })
                wf_service = netsvc.LocalService("workflow")
                wf_service.trg_validate(uid, 'mrp.procurement', proc_id, 'button_confirm', cr)

            vals={
                "date": None,
                "order_id": order.id,
                "product_id": order.product_id.id,
                "qty": order.qty,
                "uom_id": order.uom_id.id,
            }
            fin_id=self.pool.get("ac.production.finished.product").create(cr,uid,vals)
            order.write({"fin_qty_good":order.qty})
            vals={
                "name": "FIN_GOOD: "+order.name,
                "product_id": order.product_id.id,
                "product_qty": order.qty,
                "product_uom": order.uom_id.id,
                "location_id": order.routing_id.location_id.id,
                "location_dest_id": order.location_dest_id.id,
                "state": "confirmed",
                "production_id": order.id,
            }
            move_id=self.pool.get("stock.move").create(cr,uid,vals)
            order.write({"fin_move_good_id":move_id})
        return True

    def wkf_in_progress(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"in_progress"})
        for order in self.browse(cr,uid,ids):
            if not order.cons_date:
                order.write({"cons_date":time.strftime("%Y-%m-%d")})
            for cons in order.cons_products:
                cons_date=cons.date
                if not cons_date:
                    cons_date=time.strftime("%Y-%m-%d")
                    cons.write({"date": cons_date})
                if cons.move_total_id:
                    move=cons.move_total_id
                    move.write({
                        "product_qty": cons.qty_total,
                        "date": cons_date,
                    })
                else:
                    vals={
                        "name": "CONS_TOT: "+order.name,
                        "product_id": cons.product_id.id,
                        "product_qty": cons.qty_total,
                        "product_uom": cons.uom_id.id,
                        "location_id": order.location_src_id.id,
                        "location_dest_id": order.routing_id.location_id.id,
                        "date": cons_date,
                        "production_id": order.id,
                        "state": "confirmed",
                    }
                    move_id=self.pool.get("stock.move").create(cr,uid,vals)
                    cons.write({"move_total_id":move_id})
        return True

    def wkf_done(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"done"})
        for order in self.browse(cr,uid,ids):
            if order.fin_date:
                fin_date=order.fin_date
            else:
                fin_date=time.strftime("%Y-%m-%d")
                order.write({"fin_date":fin_date})
            for cons in order.cons_products:
                if cons.qty_good+cons.qty_bad!=cons.qty_total:
                    raise osv.except_osv("Error","Invalid consumed quantities")
                if cons.move_total_id:
                    move=cons.move_total_id
                    if move.product_qty!=cons.qty_total:
                        move.write({
                            "product_qty": cons.qty_total,
                            "date": fin_date,
                        })
                    self.pool.get("stock.move").action_done(cr,uid,[move.id])
                else:
                    vals={
                        "name": "CONS_TOT: "+order.name,
                        "product_id": cons.product_id.id,
                        "product_qty": cons.qty_total,
                        "product_uom": cons.uom_id.id,
                        "location_id": order.location_src_id.id,
                        "location_dest_id": order.routing_id.location_id.id,
                        "date": fin_date,
                        "production_id": order.id,
                    }
                    move_id=self.pool.get("stock.move").create(cr,uid,vals)
                    self.pool.get("stock.move").action_done(cr,uid,[move_id])
                    cons.write({"move_total_id":move_id})
                if cons.qty_good:
                    if not cons.product_id.property_stock_production:
                        raise osv.except_osv("Error","Missing production location for product: %s"%cons.product_id.name)
                    vals={
                        "name": "CONS_GOOD: "+order.name,
                        "product_id": cons.product_id.id,
                        "product_qty": cons.qty_good,
                        "product_uom": cons.uom_id.id,
                        "location_id": order.routing_id.location_id.id,
                        "location_dest_id": order.product_id.property_stock_production.id,
                        "date": fin_date,
                        "production_id": order.id,
                    }
                    move_id=self.pool.get("stock.move").create(cr,uid,vals)
                    self.pool.get("stock.move").action_done(cr,uid,[move_id])
                    cons.write({"move_good_id": move_id})
                if cons.qty_bad:
                    if not order.product_id.property_stock_scrap:
                        raise osv.except_osv("Error","Missing scrap location for product: %s"%order.product_id.name)
                    vals={
                        "name": "CONS_BAD: "+order.name,
                        "product_id": cons.product_id.id,
                        "product_qty": cons.qty_bad,
                        "product_uom": cons.uom_id.id,
                        "location_id": order.routing_id.location_id.id,
                        "location_dest_id": order.product_id.property_stock_scrap.id,
                        "date": fin_date,
                        "production_id": order.id,
                    }
                    move_id=self.pool.get("stock.move").create(cr,uid,vals)
                    self.pool.get("stock.move").action_done(cr,uid,[move_id])
                    cons.write({"move_bad_id": move_id})
            if order.fin_qty_good+order.fin_qty_bad!=order.fin_qty_total:
                raise osv.except_osv("Error","Invalid produced quantities")
            if order.fin_move_good_id:
                move=order.fin_move_good_id
                move.write({
                    "product_qty": order.fin_qty_good,
                    "date": fin_date,
                })
                self.pool.get("stock.move").action_done(cr,uid,[move.id])
            else:
                vals={
                    "name": "FIN_GOOD: "+order.name,
                    "product_id": order.product_id.id,
                    "product_qty": order.fin_qty_good,
                    "product_uom": order.uom_id.id,
                    "location_id": order.routing_id.location_id.id,
                    "location_dest_id": order.location_dest_id.id,
                    "date": fin_date,
                    "production_id": order.id,
                }
                move_id=self.pool.get("stock.move").create(cr,uid,vals)
                self.pool.get("stock.move").action_done(cr,uid,[move_id])
                order.write({"fin_move_good_id": move_id})
            if order.fin_qty_bad:
                vals={
                    "name": "FIN_BAD: "+order.name,
                    "product_id": order.product_id.id,
                    "product_qty": order.fin_qty_bad,
                    "product_uom": order.uom_id.id,
                    "location_id": order.routing_id.location_id.id,
                    "location_dest_id": order.product_id.property_stock_scrap.id,
                    "date": fin_date,
                    "production_id": order.id,
                }
                move_id=self.pool.get("stock.move").create(cr,uid,vals)
                self.pool.get("stock.move").action_done(cr,uid,[move_id])
                order.write({"fin_move_bad_id": move_id})
            vals={
                "name": "FIN_TOT: "+order.name,
                "product_id": order.product_id.id,
                "product_qty": order.fin_qty_total,
                "product_uom": order.uom_id.id,
                "location_id": order.product_id.property_stock_production.id,
                "location_dest_id": order.routing_id.location_id.id,
                "date": fin_date,
                "production_id": order.id,
            }
            move_id=self.pool.get("stock.move").create(cr,uid,vals)
            self.pool.get("stock.move").action_done(cr,uid,[move_id])
            order.write({"fin_move_total_id": move_id})
        return True

    def wkf_canceled(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"canceled"})
        for order in self.browse(cr,uid,ids):
            for cons in order.cons_products:
                for move in cons.moves:
                    move.write({"state": "cancel"})
            for fin in order.fin_products:
                for move in fin.moves:
                    move.write({"state": "cancel"})
            cr.execute('delete from ac_workcenter_operation where production_id=%s', (order.id,))
        return True

    def copy(self,cr,uid,id,default=None,context=None):
        if not default:
            default={}
        default.update({
            "name": self.pool.get('ir.sequence').get(cr,uid,'mrp.production'),
            "cons_products": [],
            "fin_products": [],
            "state": "draft",
        })
        return super(ac_production,self).copy(cr,uid,id,default,context)
ac_production()

class ac_production_consumed_product(osv.osv):
    _name="ac.production.consumed.product"

    _columns={
        "order_id": fields.many2one("ac.production","Production Order",required=True,ondelete="cascade"),
        "date": fields.date("Date"),
        "product_id": fields.many2one("product.product","Product",required=True),
        "qty_total": fields.float("Total Qty",required=True),
        "qty_good": fields.float("Good Qty"),
        "qty_bad": fields.float("Bad Qty"),
        "uom_id": fields.many2one("product.uom","UoM",required=True),
        "notes": fields.char("Remark",size=128),
        "move_total_id": fields.many2one("stock.move","Total Raw Mat Move"),
        "move_good_id": fields.many2one("stock.move","Good Raw Mat Move"),
        "move_bad_id": fields.many2one("stock.move","Bad Raw Mat Move"),
    }
    _defaults={
        "date": lambda *a: time.strftime("%Y-%m-%d"),
    }

    def onchange_qty_good(self,cr,uid,ids,qty_good,qty_total):
        if not qty_good:
            return {}
        vals={
            "qty_bad": qty_total-qty_good,
        }
        return {"value":vals}

    def onchange_qty_bad(self,cr,uid,ids,qty_bad,qty_total):
        if not qty_bad:
            return {}
        vals={
            "qty_good": qty_total-qty_bad,
        }
        return {"value":vals}
ac_production_consumed_product()

class ac_production_finished_product(osv.osv):
    _name="ac.production.finished.product"
    _columns={
        "order_id": fields.many2one("ac.production","Production Order",required=True,ondelete="cascade"),
        "date": fields.date("Date"),
        "product_id": fields.many2one("product.product","Product",required=True),
        "qty": fields.float("Quantity",required=True),
        "uom_id": fields.many2one("product.uom","UoM",required=True),
        "notes": fields.char("Remark",size=128),
    }
    _defaults={
        "date": lambda *a: time.strftime("%Y-%m-%d"),
    }
ac_production_finished_product()

class ac_production_step(osv.osv):
    _name="ac.production.step"

    def _qty_total(self,cr,uid,ids,name,arg,context={}):
        vals={}
        for work in self.browse(cr,uid,ids):
            vals[work.id]=work.qty_good+work.qty_bad
        return vals

    def name_get(self,cr,uid,ids,context={}):
        return [(step.id,str(step.sequence)) for step in self.browse(cr,uid,ids)]

    def name_search(self, cr, uid, name, args=None, operator='ilike', context=None, limit=80):
        print "name_search",name,args,operator
        dom=args or []
        try:
            seq=int(name)
            dom+=[("sequence","=",seq)]
        except:
            pass
        ids=self.search(cr,uid,dom)
        return self.name_get(cr,uid,ids)

    _columns={
        "sequence": fields.integer("Sequence",required=True),
        "order_id": fields.many2one("ac.production","Production Order",required=True,ondelete="cascade"),
        "worker_id": fields.many2one("res.users","Worker",required=True),
        "dept_id": fields.many2one("hr.department","Department",required=True),
        "next_dept_id": fields.many2one("hr.department","Next Department"),
        "receiver_id": fields.many2one("res.users","Receiver"),
        "qty_total": fields.function(_qty_total,method=True,type="float",string="Total Qty"),
        "qty_good": fields.float("Good Qty"),
        "qty_bad": fields.float("Bad Qty"),
        "date": fields.date("Date Received"),
        "operations": fields.one2many("ac.workcenter.operation","prod_step_id","Workcenter Operations"),
    }

    def _default_dept(self,cr,uid,context):
        user=self.pool.get("res.users").browse(cr,uid,uid)
        if not user.departments:
            return False
        return user.departments[0].id

    def onchange_worker(self,cr,uid,ids,worker_id):
        if not worker_id:
            return {}
        user=self.pool.get("res.users").browse(cr,uid,worker_id)
        if not user.departments:
            return False
        vals={
            "dept_id": user.departments[0].id,
        }
        return {"value": vals}

    _defaults={
        "sequence": lambda *a: 1,
        "worker_id": lambda self,cr,uid,context: uid,
        "dept_id": _default_dept,
    }
ac_production_step()

class ac_assembly(osv.osv):
    _name="ac.assembly"

    def _get_time(self,cr,uid,ids,name,args,context={}):
        vals_={}
        for order in self.browse(cr,uid,ids):
            vals={
                "time_planned": 0.0,
                "time_actual": 0.0,
            }
            for op in order.operations:
                if op.time_planned:
                    vals["time_planned"]+=op.time_planned
                if op.time_actual:
                    vals["time_actual"]+=op.time_actual
            vals_[order.id]=vals
        return vals_

    def _get_time_start(self,cr,uid,ids,name,arg,context={}):
        print "ASS _get_time_start",ids
        vals={}
        for order in self.browse(cr,uid,ids):
            if order.operations:
                time=min([op.time_start for op in order.operations])
            else:
                time=order.date_planned+" 08:00:00"
            vals[order.id]=time
        return vals

    def _get_time_stop(self,cr,uid,ids,name,arg,context={}):
        print "ASS _get_time_stop",ids
        vals={}
        for order in self.browse(cr,uid,ids):
            if order.operations:
                time=max([op.time_stop for op in order.operations])
            else:
                time=order.date_planned+" 17:00:00"
            vals[order.id]=time
        return vals

    _columns={
        "name": fields.char("Reference",size=64,required=True,select=1),
        "product_id": fields.many2one("product.product","Product",required=True,select=2,domain=[("mrp_type","=","machine")]),
        "qty": fields.float("Quantity",required=True),
        "uom_id": fields.many2one("product.uom","UoM",required=True),
        "origin": fields.char("Origin",size=64,select=1),
        "dept_id": fields.many2one("hr.department","Department"),
        "cons_pickings": fields.one2many("stock.picking","assembly_id","Consumed Products Pickings"),
        "fin_move_id": fields.many2one("stock.move","Finished Products Move"),
        "state": fields.selection([("draft","Draft"),("confirmed","Confirmed"),("in_progress","In Progress"),("done","Done"),("canceled","Canceled")],"Status",readonly=True,required=True),
        "sched_products": fields.one2many("ac.scheduled.product","assembly_id","Scheduled Products",readonly=True),
        "operations": fields.one2many("ac.workcenter.operation","assembly_id","Workcenter Operations"),
        "location_src_id": fields.many2one("stock.location","Parts Location",required=True,domain=[('usage','=','internal')]),
        "location_dest_id": fields.many2one("stock.location","Finished Machines Location",required=True,domain=[('usage','=','internal')]),
        "bom_id": fields.many2one("mrp.bom","Bill of Material"),
        "routing_id": fields.many2one("mrp.routing","Routing"),
        "serial": fields.char("Serial No.",size=64,select=2),
        "partner_id": fields.many2one("res.partner","Customer",domain=[("customer","=",True)],select=2),
        "date_planned": fields.date("Scheduled Date",required=True,select=1),
        "time_planned": fields.function(_get_time,method=True,type="float",string="Planned Time",multi="time"),
        "time_actual": fields.function(_get_time,method=True,type="float",string="Actual Time",multi="time"),
        "time_start": fields.function(_get_time_start,method=True,type="datetime",string="Start Time"),
        "time_stop": fields.function(_get_time_stop,method=True,type="datetime",string="Stop Time"),
    }
    _defaults={
        "name": lambda self,cr,uid,context: self.pool.get('ir.sequence').get(cr,uid,'ac.assembly') or '/',
        "state": lambda *a: "draft",
        "date_planned": lambda *a: time.strftime("%Y-%m-%d"),
    }
    _order="date_planned"

    def onchange_product(self,cr,uid,ids,prod_id,context={}):
        if not prod_id:
            return {}
        prod=self.pool.get("product.product").browse(cr,uid,prod_id)
        vals={
            "uom_id": prod.uom_id.id,
            "qty": 1.0,
        }
        return {"value": vals}

    def btn_compute(self,cr,uid,ids,properties=[],context={}):
        results = []
        for ass in self.browse(cr, uid, ids):
            cr.execute('delete from ac_scheduled_product where assembly_id=%s', (ass.id,))
            cr.execute('delete from ac_workcenter_operation where assembly_id=%s', (ass.id,))
            bom_point = ass.bom_id
            bom_id = ass.bom_id.id
            if not bom_point:
                bom_id = self.pool.get('mrp.bom')._bom_find(cr, uid, ass.product_id.id, ass.uom_id.id, properties)
                if bom_id:
                    bom_point = self.pool.get('mrp.bom').browse(cr, uid, bom_id)
                    routing_id = bom_point.routing_id.id or False
                    self.write(cr, uid, [ass.id], {'bom_id': bom_id, 'routing_id': routing_id})

            if not bom_id:
                raise osv.except_osv(_('Error'), _("Couldn't find bill of material for product"))

            factor = ass.qty * ass.uom_id.factor / bom_point.product_uom.factor
            res = self.pool.get('mrp.bom')._bom_explode(cr, uid, bom_point, factor / bom_point.product_qty, properties)
            results = res[0]
            results2 = res[1]
            for line in results:
                vals={
                    "assembly_id": ass.id,
                    "product_id": line["product_id"],
                    "qty": line["product_qty"],
                    "uom_id": line["product_uom"],
                    "bom_id": line["bom_id"],
                }
                self.pool.get('ac.scheduled.product').create(cr, uid, vals)
            for line in results2:
                vals={
                    "assembly_id": ass.id,
                    "sequence": line["sequence"],
                    "name": line["name"],
                    "routing_id": line["routing_id"],
                    "workcenter_id": line["workcenter_id"],
                    "time_planned": line["hour"]*60.0,
                    "date": ass.date_planned,
                }
                self.pool.get('ac.workcenter.operation').create(cr, uid, vals)
        return True

    def wkf_draft(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"draft"})
        return True

    def wkf_confirmed(self,cr,uid,ids,context={}):
        for order in self.browse(cr,uid,ids):
            if not order.routing_id or not order.routing_id.location_id:
                raise osv.except_osv("Error","Missing routing production location")
            pick_ids={}
            for prod in order.sched_products:
                pick_id=pick_ids.get(prod.bom_id.id)
                if not pick_id:
                    vals={
                        "origin": order.name+" [%s]"%prod.bom_id.name,
                        "assembly_id": order.id,
                        "type": "internal",
                    }
                    pick_id=self.pool.get("stock.picking").create(cr,uid,vals)
                    pick_ids[prod.bom_id.id]=pick_id
                vals={
                    "picking_id": pick_id,
                    "name": "CONS: "+order.name,
                    "product_id": prod.product_id.id,
                    "product_qty": prod.qty,
                    "product_uom": prod.uom_id.id,
                    "location_id": order.location_src_id.id,
                    "location_dest_id": order.routing_id.location_id.id,
                    "state": "confirmed",
                    "assembly_id": order.id,
                }
                move_id=self.pool.get("stock.move").create(cr,uid,vals)
                proc_id = self.pool.get('mrp.procurement').create(cr, uid, {
                    'name': (order.origin or '').split(':')[0] + ':' + order.name,
                    'origin': (order.origin or '').split(':')[0] + ':' + order.name,
                    'date_planned': order.date_planned,
                    'product_id': prod.product_id.id,
                    'product_qty': prod.qty,
                    'product_uom': prod.uom_id.id,
                    'location_id': order.location_src_id.id,
                    'procure_method': prod.product_id.procure_method,
                    'move_id': move_id,
                })
                wf_service = netsvc.LocalService("workflow")
                wf_service.trg_validate(uid, 'mrp.procurement', proc_id, 'button_confirm', cr)
            for pick_id in pick_ids.values():
                wf_service = netsvc.LocalService("workflow")
                wf_service.trg_validate(uid, 'stock.picking', pick_id, 'button_confirm', cr)
            vals={
                "name": "FIN: "+order.name,
                "product_id": order.product_id.id,
                "product_qty": order.qty,
                "product_uom": order.uom_id.id,
                "location_id": order.product_id.property_stock_production.id,
                "location_dest_id": order.location_dest_id.id,
                "state": "confirmed",
                "assembly_id": order.id,
            }
            move_id=self.pool.get("stock.move").create(cr,uid,vals)
            order.write({"fin_move_id":move_id})
        self.write(cr,uid,ids,{"state":"confirmed"})
        return True

    def wkf_in_progress(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"in_progress"})
        return True

    def wkf_done(self,cr,uid,ids,context={}):
        for order in self.browse(cr,uid,ids):
            move=order.fin_move_id
            move.write({"state":"done"})
            move_ids=self.pool.get("stock.move").search(cr,uid,[("assembly_id","=",order.id),("name","like","CONS")])
            for move in self.pool.get("stock.move").browse(cr,uid,move_ids):
                if move.state=="done":
                    move.copy({
                        "location_id": order.routing_id.location_id.id,
                        "location_dest_id": move.product_id.property_stock_production.id,
                        "state": "confirmed",
                    })
                    move.action_done()
        self.write(cr,uid,ids,{"state":"done"})
        return True

    def wkf_canceled(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"canceled"})
        for order in self.browse(cr,uid,ids):
            for pick in order.cons_pickings:
                pick.action_cancel()
            for move in order.fin_moves:
                move.write({"state": "cancel"})
            cr.execute('delete from ac_workcenter_operation where assembly_id=%s', (order.id,))
        return True

    def copy(self,cr,uid,id,default=None,context=None):
        if not default:
            default={}
        default.update({
            "name": self.pool.get('ir.sequence').get(cr,uid,'ac.assembly'),
            "cons_pickings": [],
            "fin_moves": [],
            "state": "draft",
        })
        return super(ac_assembly,self).copy(cr,uid,id,default,context)
ac_assembly()

class stock_move(osv.osv):
    _inherit="stock.move"
    _columns={
        "production_id": fields.many2one("ac.production","Production"),
        "assembly_id": fields.many2one("ac.assembly","Assembly"),
    }
stock_move()

class stock_picking(osv.osv):
    _inherit="stock.picking"
    _columns={
        "production_id": fields.many2one("ac.production","Production"),
        "assembly_id": fields.many2one("ac.assembly","Assembly"),
    }
stock_picking()

class ac_scheduled_product(osv.osv):
    _name="ac.scheduled.product"
    _columns={
        "production_id": fields.many2one("ac.production","Production Order",ondelete="cascade"),
        "assembly_id": fields.many2one("ac.assembly","Assembly Order",ondelete="cascade"),
        "product_id": fields.many2one("product.product","Product",required=True),
        "qty": fields.float("Required Qty",required=True),
        "uom_id": fields.many2one("product.uom","UoM",required=True),
        "qty_available": fields.related("product_id","qty_available",type="float",string="In Stock"),
        "bom_id": fields.many2one("mrp.bom","BoM"),
    }
ac_scheduled_product()

class ac_workcenter_operation(osv.osv):
    _name="ac.workcenter.operation"

    def _get_time_actual(self,cr,uid,ids,name,arg,context={}):
        vals={}
        for op in self.browse(cr,uid,ids):
            if op.start and op.start!="00:00:00" and op.stop and op.stop!="00:00:00":
                start_=datetime.datetime.strptime(op.time_start,"%Y-%m-%d %H:%M:%S")
                stop_=datetime.datetime.strptime(op.time_stop,"%Y-%m-%d %H:%M:%S")
                if start_>stop_:
                    vals[op.id]=False
                else:
                    vals[op.id]=((stop_-start_).seconds+59)/60
            else:
                vals[op.id]=False
        return vals

    def _get_time_start(self,cr,uid,ids,name,arg,context={}):
        print "OP _get_time_start",ids
        vals={}
        for op in self.browse(cr,uid,ids):
            if op.start and op.start!="00:00:00":
                vals[op.id]=op.date+" "+op.start
            else:
                vals[op.id]=op.date+" 08:00:00"
        return vals

    def _get_time_stop(self,cr,uid,ids,name,arg,context={}):
        print "OP_get_time_stop",ids
        vals={}
        for op in self.browse(cr,uid,ids):
            print "stop",op.stop,"xxx"
            if op.stop and op.stop!="00:00:00":
                vals[op.id]=op.date+" "+op.stop
            else:
                print "start",op.time_start
                print "time",op.time_planned
                start_=datetime.datetime.strptime(op.time_start,"%Y-%m-%d %H:%M:%S")
                stop_=start_+datetime.timedelta(minutes=op.time_planned or 0.0)
                vals[op.id]=stop_.strftime("%Y-%m-%d %H:%M:%S")
        return vals

    def _set_time_start(self,cr,uid,ids,name,value,arg,context):
        print "_set_time_start",ids
        if type(ids)!=type([]):
            ids=[ids]
        for op in self.browse(cr,uid,ids):
            op.write({"date":value[:10],"start": value[11:]})
            pass
        return True

    def _set_time_stop(self,cr,uid,ids,name,value,arg,context):
        print "_set_time_stop",ids
        if type(ids)!=type([]):
            ids=[ids]
        for op in self.browse(cr,uid,ids):
            op.write({"date":value[:10],"stop": value[11:]})
        return True

    def _product(self,cr,uid,ids,name,arg,context={}):
        vals={}
        for op in self.browse(cr,uid,ids):
            if op.production_id:
                vals[op.id]=op.production_id.product_id.id
            elif op.assembly_id:
                vals[op.id]=op.assembly_id.product_id.id
            else:
                vals[op.id]=False
        return vals

    _columns={
        "workcenter_id": fields.many2one("mrp.workcenter","Workcenter",required=True,select=1),
        "date": fields.date("Date",required=True,select=1),
        "start": fields.time("Start"),
        "stop": fields.time("Stop"),
        "name": fields.char("Description",size=64,required=True,select=2),
        "time_planned": fields.float("Planned Time"),
        "time_actual": fields.function(_get_time_actual,method=True,type="float",string="Actual Time",store=True),
        "time_start": fields.function(_get_time_start,method=True,type="datetime",string="Start Time",fnct_inv=_set_time_start),
        "time_stop": fields.function(_get_time_stop,method=True,type="datetime",string="Stop Time",fnct_inv=_set_time_stop),
        "product_id": fields.function(_product,method=True,type="many2one",relation="product.product",string="Product"),
        # used for part production
        "production_id": fields.many2one("ac.production","Production",ondelete="cascade",select=1),
        "qty": fields.float("Qty"),
        "prod_step_id": fields.many2one("ac.production.step","Production Step"),
        # used for machine assembly
        "assembly_id": fields.many2one("ac.assembly","Assembly",ondelete="cascade",select=1),
        "routing_id": fields.many2one("mrp.routing","Routing"),
        "worker_id": fields.many2one("res.users","Worker"),
        "remark": fields.char("Remark",size=64),
    }

    _defaults={
        "date": lambda *a: time.strftime("%Y-%m-%d"),
    }
ac_workcenter_operation()

class product_product(osv.osv):
    _inherit="product.product"
    _columns={
        "mrp_type": fields.selection([("part","Part Production"),("machine","Machine Assembly")],"Production Type"),
        'property_stock_scrap': fields.property(
            'stock.location',
            type='many2one',
            relation='stock.location',
            string="Scrap Location",
            method=True,
            view_load=True),
    }
product_product()

def rounding(f, r):
    if not r:
        return f
    return round(f / r) * r

class mrp_bom(osv.osv):
    _inherit="mrp.bom"

    def _bom_explode(self, cr, uid, bom, factor, properties, addthis=False, level=0):
        print "_bom_explode",bom.product_id.name,bom.type,factor,addthis,level
        factor = factor / (bom.product_efficiency or 1.0)
        factor = rounding(factor, bom.product_rounding)
        if factor<bom.product_rounding:
            factor = bom.product_rounding
        result = []
        result2 = []
        phantom=False
        if bom.type=='phantom' and not bom.bom_lines:
            newbom = self._bom_find(cr, uid, bom.product_id.id, bom.product_uom.id, properties)
            if newbom:
                res = self._bom_explode(cr, uid, self.browse(cr, uid, [newbom])[0], factor*bom.product_qty, properties, addthis=True, level=level+10)
                result = result + res[0]
                result2 = result2 + res[1]
                phantom=True
            else:
                phantom=False
        if not phantom:
            if addthis and not bom.bom_lines:
                result.append(
                {
                    'name': bom.product_id.name,
                    'product_id': bom.product_id.id,
                    'product_qty': bom.product_qty * factor,
                    'product_uom': bom.product_uom.id,
                    'product_uos_qty': bom.product_uos and bom.product_uos_qty * factor or False,
                    'product_uos': bom.product_uos and bom.product_uos.id or False,
                    "bom_id": bom.bom_id.id,
                })
            if bom.routing_id:
                for wc_use in bom.routing_id.workcenter_lines:
                    wc = wc_use.workcenter_id
                    d, m = divmod(factor, wc_use.workcenter_id.capacity_per_cycle)
                    mult = (d + (m and 1.0 or 0.0))
                    cycle = mult * wc_use.cycle_nbr
                    result2.append({
                        'name': wc_use.name,
                        'workcenter_id': wc.id,
                        'sequence': level+(wc_use.sequence or 0),
                        'cycle': cycle,
                        'hour': float(wc_use.hour_nbr*mult + (wc.time_start+wc.time_stop+cycle*wc.time_cycle) * (wc.time_efficiency or 1.0)),
                        "routing_id": bom.routing_id.id,
                    })
            for bom2 in bom.bom_lines:
                res = self._bom_explode(cr, uid, bom2, factor, properties, addthis=True, level=level+10)
                result = result + res[0]
                result2 = result2 + res[1]
        return result, result2
mrp_bom()

class mrp_procurement(osv.osv):
    _inherit="mrp.procurement"

    def check_product_part(self, cr, uid, ids):
        for procurement in self.browse(cr, uid, ids):
            if procurement.product_id.type in ('product', 'consu'):
                if procurement.product_id.mrp_type=="part":
                    return True
                if procurement.product_id.mrp_type not in ("part","machine"):
                    cr.execute('update mrp_procurement set message=%s where id=%s', (_('Invalid production type!'), procurement.id))
        return False

    def check_product_machine(self, cr, uid, ids):
        for procurement in self.browse(cr, uid, ids):
            if procurement.product_id.type in ('product', 'consu'):
                if procurement.product_id.mrp_type=="machine":
                    return True
                if procurement.product_id.mrp_type not in ("part","machine"):
                    cr.execute('update mrp_procurement set message=%s where id=%s', (_('Invalid production type!'), procurement.id))
        return False

    def action_produce_assign_product(self, cr, uid, ids, context={}):
        produce_id = False
        company = self.pool.get('res.users').browse(cr, uid, uid, context).company_id
        for procurement in self.browse(cr, uid, ids):
            res_id = procurement.move_id.id
            loc_id = procurement.location_id.id
            newdate = DateTime.strptime(procurement.date_planned, '%Y-%m-%d %H:%M:%S') - DateTime.RelativeDateTime(days=procurement.product_id.product_tmpl_id.produce_delay or 0.0)
            newdate = newdate - DateTime.RelativeDateTime(days=company.manufacturing_lead)
            order_id = self.pool.get("ac.production").create(cr, uid, {
                'origin': procurement.origin,
                'product_id': procurement.product_id.id,
                'qty': procurement.product_qty,
                'uom_id': procurement.product_uom.id,
                #'location_src_id': procurement.location_id.id, # should have default
                'location_dest_id': procurement.location_id.id,
                'bom_id': procurement.bom_id and procurement.bom_id.id or False,
                'date_planned': newdate.strftime('%Y-%m-%d %H:%M:%S'),
            })
            self.write(cr, uid, [procurement.id], {'state':'running'})
            bom_result = self.pool.get("ac.production").btn_compute(cr, uid,
                    [order_id], properties=[x.id for x in procurement.property_ids])
            wf_service = netsvc.LocalService("workflow")
            wf_service.trg_validate(uid, "ac.production", order_id, 'btn_confirm', cr)
            self.pool.get('stock.move').write(cr, uid, [res_id],
                    {'location_id':procurement.location_id.id})
        return order_id

    def action_assemble(self, cr, uid, ids, context={}):
        produce_id = False
        company = self.pool.get('res.users').browse(cr, uid, uid, context).company_id
        for procurement in self.browse(cr, uid, ids):
            res_id = procurement.move_id.id
            loc_id = procurement.location_id.id
            newdate = DateTime.strptime(procurement.date_planned, '%Y-%m-%d %H:%M:%S') - DateTime.RelativeDateTime(days=procurement.product_id.product_tmpl_id.produce_delay or 0.0)
            newdate = newdate - DateTime.RelativeDateTime(days=company.manufacturing_lead)
            order_id = self.pool.get("ac.assembly").create(cr, uid, {
                'origin': procurement.origin,
                'product_id': procurement.product_id.id,
                'qty': procurement.product_qty,
                'uom_id': procurement.product_uom.id,
                #'location_src_id': procurement.location_id.id, # should have default
                'location_dest_id': procurement.location_id.id,
                'bom_id': procurement.bom_id and procurement.bom_id.id or False,
                'date_planned': newdate.strftime('%Y-%m-%d %H:%M:%S'),
            })
            self.write(cr, uid, [procurement.id], {'state':'running'})
            bom_result = self.pool.get("ac.assembly").btn_compute(cr, uid,
                    [order_id], properties=[x.id for x in procurement.property_ids])
            wf_service = netsvc.LocalService("workflow")
            wf_service.trg_validate(uid, "ac.assembly", order_id, 'btn_confirm', cr)
            self.pool.get('stock.move').write(cr, uid, [res_id],
                    {'location_id':procurement.location_id.id})
        return order_id
mrp_procurement()

class res_users(osv.osv):
    _inherit="res.users"
    _columns={
        "departments": fields.many2many("hr.department","hr_department_user_rel","user_id","department_id","Departments"),
    }
res_users()

class ac_report_workcenter_load(osv.osv):
    _name="ac.report.workcenter.load"
    _description="Workcenter Load"
    _auto = False
    _log_access = False
    _columns = {
        'name': fields.char('Week', size=64, required=True),
        'workcenter_id': fields.many2one('mrp.workcenter', 'Workcenter', required=True),
        'time_planned': fields.float('Planned Time'),
        'time_actual': fields.float('Actual Time'),
    }

    def init(self, cr):
        cr.execute("""
            create or replace view ac_report_workcenter_load as (
                SELECT
                    min(op.id) as id,
                    to_char(op.date,'YYYY:IW') as name,
                    SUM(op.time_planned) AS time_planned,
                    SUM(op.time_actual) AS time_actual,
                    op.workcenter_id as workcenter_id
                FROM
                    ac_workcenter_operation op
                GROUP BY
                    op.workcenter_id,
                    to_char(op.date,'YYYY:IW')
            )""")
ac_report_workcenter_load()

